# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from __future__ import annotations

from typing import List, Dict, Optional, TYPE_CHECKING

from ethtx.models.base_model import BaseModel

if TYPE_CHECKING:
    from ethtx.providers.semantic_providers import ISemanticsDatabase


class TransformationSemantics(BaseModel):
    transformed_name: Optional[str]
    transformed_type: Optional[str]
    transformation: str = ""


class ParameterSemantics(BaseModel):
    parameter_name: str
    parameter_type: str
    components: list[ParameterSemantics] = []
    indexed: bool = False
    dynamic: bool = False


ParameterSemantics.update_forward_refs()


class EventSemantics(BaseModel):
    signature: str
    anonymous: bool
    name: str
    parameters: List[ParameterSemantics]


class FunctionSemantics(BaseModel):
    signature: str
    name: str
    inputs: List[ParameterSemantics]
    outputs: List[ParameterSemantics] = []


class SignatureArg(BaseModel):
    name: str
    type: str


class Signature(BaseModel):
    signature_hash: str
    name: str
    args: List[SignatureArg]
    count: int = 1
    tuple: bool = False
    guessed: bool = False


class ERC20Semantics(BaseModel):
    name: str
    symbol: str
    decimals: int


class ContractSemantics(BaseModel):
    code_hash: str
    name: str
    events: Dict[str, EventSemantics] = {}
    functions: Dict[str, FunctionSemantics] = {}
    transformations: Dict[str, Dict[str, TransformationSemantics]] = {}


class AddressSemantics(BaseModel):
    chain_id: str
    address: str
    name: str
    is_contract: bool
    contract: ContractSemantics
    standard: Optional[str]
    erc20: Optional[ERC20Semantics]

    class Config:
        allow_mutation = True

    @staticmethod
    def from_mongo_record(
        raw_address_semantics: Dict, database: "ISemanticsDatabase"
    ) -> "AddressSemantics":

        ZERO_HASH = "0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470"

        def decode_parameter(_parameter):
            components_semantics = []
            if "components" in _parameter:
                for component in _parameter["components"]:
                    components_semantics.append(decode_parameter(component))

            decoded_parameter = ParameterSemantics(
                parameter_name=_parameter["parameter_name"],
                parameter_type=_parameter["parameter_type"],
                components=components_semantics,
                indexed=_parameter["indexed"],
                dynamic=_parameter["dynamic"],
            )

            return decoded_parameter

        if raw_address_semantics.get("erc20"):
            erc20_semantics = ERC20Semantics(
                name=raw_address_semantics["erc20"]["name"],
                symbol=raw_address_semantics["erc20"]["symbol"],
                decimals=raw_address_semantics["erc20"]["decimals"],
            )
        else:
            erc20_semantics = None

        if raw_address_semantics["contract"] == ZERO_HASH:
            contract_semantics = ContractSemantics(
                code_hash=raw_address_semantics["contract"], name="EOA"
            )

        else:

            raw_contract_semantics = database.get_contract_semantics(
                raw_address_semantics["contract"]
            )
            events = {}

            for signature, event in raw_contract_semantics["events"].items():

                parameters_semantics = []
                for parameter in event["parameters"]:
                    parameters_semantics.append(decode_parameter(parameter))

                events[signature] = EventSemantics(
                    signature=signature,
                    anonymous=event["anonymous"],
                    name=event["name"],
                    parameters=parameters_semantics,
                )

            functions = {}
            for signature, function in raw_contract_semantics["functions"].items():

                inputs_semantics = []
                for parameter in function["inputs"]:
                    inputs_semantics.append(decode_parameter(parameter))
                outputs_semantics = []
                for parameter in function["outputs"]:
                    outputs_semantics.append(decode_parameter(parameter))

                functions[signature] = FunctionSemantics(
                    signature=signature,
                    name=function["name"],
                    inputs=inputs_semantics,
                    outputs=outputs_semantics,
                )

            transformations = {}
            for signature, parameters_transformations in raw_contract_semantics[
                "transformations"
            ].items():
                transformations[signature] = {}
                for parameter, transformation in parameters_transformations.items():
                    transformations[signature][parameter] = TransformationSemantics(
                        transformed_name=transformation["transformed_name"],
                        transformed_type=transformation["transformed_type"],
                        transformation=transformation["transformation"],
                    )

            contract_semantics = ContractSemantics(
                code_hash=raw_contract_semantics["code_hash"],
                name=raw_contract_semantics["name"],
                events=events,
                functions=functions,
                transformations=transformations,
            )

        address = raw_address_semantics.get("address")
        chain_id = raw_address_semantics.get("chain_id")
        name = raw_address_semantics.get("name", address)

        address_semantics = AddressSemantics(
            chain_id=chain_id,
            address=address,
            name=name,
            is_contract=raw_address_semantics["is_contract"],
            contract=contract_semantics,
            standard=raw_address_semantics["standard"],
            erc20=erc20_semantics,
        )

        return address_semantics



from ethtx.models.semantics_model import ParameterSemantics


ERRORS = {
    '0x08c379a0': {
        'name': 'Error',
        'abi': [
            ParameterSemantics(
                parameter_name="Error",
                parameter_type="string",
                components=[],
                indexed=False,
                dynamic=True,
            )
        ]
    },
    '0xfdb6ca8d': {
        'name': 'OrderStatusError',
        'abi': [
            ParameterSemantics(
                parameter_name="orderHash",
                parameter_type="bytes32",
            ),
            ParameterSemantics(
                parameter_name="orderStatus",
                parameter_type="uint8"
            )
        ]
    },
    '0x990174d2': {
        'name': 'IncompleteTransformERC20Error',
        'abi':
            [
                ParameterSemantics(
                    parameter_name="outputToken",
                    parameter_type="address"
                ),
                ParameterSemantics(
                    parameter_name="outputTokenAmount",
                    parameter_type="uint256"
                ),
                ParameterSemantics(
                    parameter_name="minOutputTokenAmount",
                    parameter_type="uint256"
                )
            ]
    },
    '0x4678472b': {
        'name': 'AssetProxyTransferError',
        'abi':
            [
                ParameterSemantics(
                    parameter_name="orderHash",
                    parameter_type="bytes32"
                ),
                ParameterSemantics(
                    parameter_name="assetData",
                    parameter_type="bytes"
                ),
                ParameterSemantics(
                    parameter_name="errorData",
                    parameter_type="bytes"
                )
            ]
    },
    '0x87cb1e75': {
        'name': 'PayProtocolFeeError',
        'abi':
            [
                ParameterSemantics(
                    parameter_name="orderHash",
                    parameter_type="bytes32"
                ),
                ParameterSemantics(
                    parameter_name="protocolFee",
                    parameter_type="uint256"
                ),
                ParameterSemantics(
                    parameter_name="makerAddress",
                    parameter_type="address"
                ),
                ParameterSemantics(
                    parameter_name="takerAddress",
                    parameter_type="address"
                ),
                ParameterSemantics(
                    parameter_name="errorData",
                    parameter_type="bytes"
                )
            ]
    },
    '0x339f3de2': {
        'name': 'RoundingError',
        'abi':
            [
                ParameterSemantics(
                    parameter_name="numerator",
                    parameter_type='uint256'
                ),
                ParameterSemantics(
                    parameter_name="denominator",
                    parameter_type='uint256'
                ),
                ParameterSemantics(
                    parameter_name="target",
                    parameter_type='uint256'
                ),
            ]
    },
    '0x18e4b141': {
        'name': 'IncompleteFillError',
        'abi': [
            ParameterSemantics(
                parameter_name='errorCode',
                parameter_type='uint8'
            ),
            ParameterSemantics(
                parameter_name='expectedAssetFillAmount',
                parameter_type='uint256'
            ),
            ParameterSemantics(
                parameter_name='actualAssetFillAmount',
                parameter_type='uint256'
            )
        ]
    }
}


from mongomock import Database

from ethtx.providers.semantic_providers.const import MongoCollections


class TestMongoSemanticsDatabase:
    def test_no_address_semantics(self, mongo_semantics_database):
        sema = mongo_semantics_database.get_address_semantics("mainnet", "not_existing")
        assert not sema

    def test_save_and_get_contract_semantic(
        self, mongo_semantics_database, mongo_db: Database
    ):
        code_hash = "test_code_hash"
        contract_data = {"code_hash": code_hash, "chain_id": "mainnet"}

        try:
            assert (
                0
                == mongo_db.get_collection(
                    MongoCollections.CONTRACTS
                ).estimated_document_count()
            )
            mongo_semantics_database.insert_contract(contract_data)
            assert (
                1
                == mongo_db.get_collection(
                    MongoCollections.CONTRACTS
                ).estimated_document_count()
            )
            contract_from_db = mongo_semantics_database.get_contract_semantics(
                code_hash
            )
            assert contract_from_db == contract_data
            assert mongo_db.list_collection_names() == [MongoCollections.CONTRACTS]
        finally:
            mongo_db.drop_collection(MongoCollections.CONTRACTS)

    def test_save_and_get_address_semantic(self, mongo_db, mongo_semantics_database):
        address = "test_address"
        address_data = {
            "address": address,
            "chain_id": "mainnet",
            "erc20": {
                "name": "test_name",
                "symbol": "test_symbol",
                "decimals": "test_decimal",
            },
            "contract": "test_contract",
            "name": "test_contract_name",
            "is_contract": False,
            "standard": False,
        }

        try:
            assert (
                0
                == mongo_db.get_collection(
                    MongoCollections.ADDRESSES
                ).estimated_document_count()
            )
            mongo_semantics_database.insert_address(address_data)
            assert (
                1
                == mongo_db.get_collection(
                    MongoCollections.ADDRESSES
                ).estimated_document_count()
            )
            address_from_db = mongo_semantics_database.get_address_semantics(
                "mainnet", address
            )
            assert address_from_db == address_data
            assert mongo_db.list_collection_names() == [MongoCollections.ADDRESSES]
        finally:
            mongo_db.drop_collection(MongoCollections.ADDRESSES)



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from __future__ import annotations

from datetime import datetime
from typing import List, Any, Optional
from decimal import Decimal, getcontext

from pydantic import validator

from ethtx.models.base_model import BaseModel
from ethtx.models.objects_model import BlockMetadata
from ethtx.models.semantics_model import AddressSemantics, ERC20Semantics


class AddressInfo(BaseModel):
    address: Optional[str]
    name: str
    badge: Optional[str]


class DecodedTransactionMetadata(BaseModel):
    chain_id: Optional[str]
    tx_hash: str
    block_number: Optional[int]
    block_hash: Optional[str]
    timestamp: Optional[datetime]
    gas_price: Optional[int]
    from_address: Optional[str]
    to_address: Optional[str]
    sender: Optional[AddressInfo]
    receiver: Optional[AddressInfo]
    tx_index: int
    tx_value: int
    gas_limit: int
    gas_used: int
    success: bool


class Argument(BaseModel):
    name: str
    type: str
    value: Any

    @validator("value")
    def decimal_conv(cls, v: Any) -> Any:
        """Method dealing with the case of large int and float by handling them as Decimal. Avoids loss of precision
        in digits.
        """
        if isinstance(v, int) or isinstance(v, float):
            getcontext().prec = 256
            return Decimal(v)
        return v


class DecodedEvent(BaseModel):
    chain_id: str
    tx_hash: str
    timestamp: datetime
    contract: AddressInfo
    index: Optional[int]
    call_id: Optional[str]
    event_signature: Optional[str]
    event_name: str
    parameters: List[Argument]
    event_guessed: bool = False


class DecodedCall(BaseModel):
    chain_id: str
    timestamp: datetime
    tx_hash: str
    call_id: Optional[str]
    call_type: str
    from_address: AddressInfo
    to_address: Optional[AddressInfo]
    value: Decimal
    function_signature: str
    function_name: str
    arguments: List[Argument]
    outputs: List[Argument]
    gas_used: Optional[int]
    error: Optional[str]
    status: bool
    indent: int
    subcalls: List[DecodedCall] = []
    function_guessed: bool = False


class DecodedTransfer(BaseModel):
    from_address: AddressInfo
    to_address: AddressInfo
    token_address: Optional[str]
    token_symbol: str
    token_standard: Optional[str]
    value: Decimal


class DecodedBalance(BaseModel):
    holder: AddressInfo
    tokens: List[dict]


class DecodedTransaction(BaseModel):
    block_metadata: BlockMetadata
    metadata: DecodedTransactionMetadata
    events: List[DecodedEvent]
    calls: Optional[DecodedCall]
    transfers: List[DecodedTransfer]
    balances: List[DecodedBalance]
    status: bool = False


class Proxy(BaseModel):
    address: str
    name: str
    type: str
    semantics: Optional[List[AddressSemantics]]
    token: Optional[ERC20Semantics]



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.


from ethtx.models.semantics_model import FunctionSemantics, ParameterSemantics

precompiles = {
    1: FunctionSemantics(
        signature="",
        name="ecrecover",
        inputs=[
            ParameterSemantics(parameter_name="hash", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="v", parameter_type="bytes8"),
            ParameterSemantics(parameter_name="r", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="s", parameter_type="bytes32"),
        ],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="address")],
    ),
    2: FunctionSemantics(
        signature="",
        name="sha256",
        inputs=[ParameterSemantics(parameter_name="data", parameter_type="raw")],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32")],
    ),
    3: FunctionSemantics(
        signature="",
        name="ripemd160",
        inputs=[ParameterSemantics(parameter_name="data", parameter_type="raw")],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32")],
    ),
    4: FunctionSemantics(
        signature="",
        name="datacopy",
        inputs=[ParameterSemantics(parameter_name="data", parameter_type="raw")],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="raw")],
    ),
    5: FunctionSemantics(
        signature="",
        name="bigModExp",
        inputs=[
            ParameterSemantics(parameter_name="base", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="exp", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="mod", parameter_type="bytes32"),
        ],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32")],
    ),
    6: FunctionSemantics(
        signature="",
        name="bn256Add",
        inputs=[
            ParameterSemantics(parameter_name="ax", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="ay", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="bx", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="by", parameter_type="bytes32"),
        ],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32[2]")],
    ),
    7: FunctionSemantics(
        signature="",
        name="bn256ScalarMul",
        inputs=[
            ParameterSemantics(parameter_name="x", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="y", parameter_type="bytes32"),
            ParameterSemantics(parameter_name="scalar", parameter_type="bytes32"),
        ],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32[2]")],
    ),
    8: FunctionSemantics(
        signature="",
        name="bn256Pairing",
        inputs=[ParameterSemantics(parameter_name="input", parameter_type="raw")],
        outputs=[ParameterSemantics(parameter_name="", parameter_type="bytes32")],
    ),
}



from ethtx.models.semantics_model import (
    TransformationSemantics,
    ParameterSemantics,
    EventSemantics,
    FunctionSemantics,
    SignatureArg,
    Signature,
    ERC20Semantics,
    ContractSemantics,
    AddressSemantics,
)
from tests.models.mock import SemanticModelMock


class TestSemanticsModels:
    def test_transformation_semantics(self):
        ts = TransformationSemantics()

        assert ts.transformed_name is None
        assert ts.transformed_type is None
        assert ts.transformation == ""

    def test_parameter_semantics(self):
        ps = ParameterSemantics(parameter_name="name", parameter_type="type")

        assert ps.parameter_name == "name"
        assert ps.parameter_type == "type"
        assert ps.components == []
        assert not ps.indexed
        assert not ps.dynamic

    def test_event_semantics(self):
        es = EventSemantics(
            signature="0x",
            anonymous=True,
            name="name",
            parameters=[SemanticModelMock.PARAMETER_SEMANTICS],
        )

        assert es.signature == "0x"
        assert es.anonymous
        assert es.name == "name"
        assert es.parameters == [SemanticModelMock.PARAMETER_SEMANTICS]

    def test_function_semantics(self):
        fs = FunctionSemantics(
            signature="0x",
            name="name",
            inputs=[SemanticModelMock.PARAMETER_SEMANTICS],
            outputs=[SemanticModelMock.PARAMETER_SEMANTICS],
        )

        assert fs.signature == "0x"
        assert fs.name == "name"
        assert fs.inputs == [SemanticModelMock.PARAMETER_SEMANTICS]
        assert fs.outputs == [SemanticModelMock.PARAMETER_SEMANTICS]

    def test_signature_args(self):
        sa = SignatureArg(name="name", type="type")

        assert sa.name == "name"
        assert sa.type == "type"

    def test_signature(self):
        sa = SignatureArg(name="name", type="type")
        s = Signature(signature_hash="0x", name="name", args=[sa])

        assert s.signature_hash == "0x"
        assert s.name == "name"
        assert s.args == [sa]
        assert s.count == 1
        assert not s.tuple
        assert not s.guessed

    def test_erc20_semantics(self):
        es = ERC20Semantics(name="name", symbol="symbol", decimals=18)

        assert es.name == "name"
        assert es.symbol == "symbol"
        assert es.decimals == 18

    def test_contract_semantics(self):
        cs = ContractSemantics(code_hash="0x", name="name")

        assert cs.code_hash == "0x"
        assert cs.name == "name"
        assert cs.events == {}
        assert cs.functions == {}
        assert cs.transformations == {}

    def test_address_semantics(self):
        ads = AddressSemantics(
            chain_id="mainnet",
            address="0x",
            name="name",
            is_contract=True,
            contract=SemanticModelMock.CONTRACT_SEMANTICS,
        )
        assert ads.chain_id == "mainnet"
        assert ads.address == "0x"
        assert ads.name == "name"
        assert ads.is_contract
        assert ads.contract == SemanticModelMock.CONTRACT_SEMANTICS
        assert ads.standard is None
        assert ads.erc20 is None



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

import logging
from typing import Optional, List, Tuple

from ethtx.models.semantics_model import (
    FunctionSemantics,
    ParameterSemantics,
    Signature,
    SignatureArg,
)
from ethtx.providers.semantic_providers import SemanticsRepository
from ethtx.providers.signature_provider import SignatureProvider, FourByteProvider

log = logging.getLogger(__name__)


def decode_function_abi_with_repository(
    signature: str, repository: SemanticsRepository
) -> Tuple[bool, Optional[FunctionSemantics]]:
    signature_obj = repository.get_most_used_signature(signature_hash=signature)
    if signature_obj:
        log.info(
            "Function (signature: %s, name: %s) guessed from SemanticsRepository.",
            signature_obj.signature_hash,
            signature_obj.name,
        )
        function_semantics = FunctionSemantics(
            signature=signature,
            name=signature_obj.name,
            inputs=_prepare_parameter_semantics(
                signature_obj.args, signature_obj.tuple, unknown=False
            ),
        )

        return signature_obj.guessed, function_semantics

    return False, None


def decode_function_abi_with_external_source(
    signature: str,
    _provider: Optional[SignatureProvider] = FourByteProvider,
) -> List[Optional[Tuple[bool, FunctionSemantics]]]:
    functions = _provider.get_function(signature=signature)

    return [
        (
            True,
            FunctionSemantics(
                signature=signature,
                name=func.get("name"),
                inputs=_prepare_parameter_semantics(
                    func.get("args"),
                    isinstance(func.get("args"), tuple),
                    unknown=True,
                ),
            ),
        )
        for func in functions
        if func
    ]


def upsert_guessed_function_semantics(
    signature: str,
    function_semantics: FunctionSemantics,
    repository: SemanticsRepository,
) -> None:
    log.info(
        "Function (signature: %s, name: %s) guessed from external source.",
        signature,
        function_semantics.name,
    )

    repository.update_or_insert_signature(
        signature=Signature(
            signature_hash=signature,
            name=function_semantics.name,
            args=[
                SignatureArg(name=f"arg_{i}", type=arg.parameter_type)
                for i, arg in enumerate(function_semantics.inputs)
            ],
            tuple=isinstance(function_semantics.inputs, tuple),
            guessed=True,
        )
    )


def decode_event_abi_name_with_external_source(
    signature: str, _provider: Optional[SignatureProvider] = FourByteProvider
) -> Tuple[bool, str]:
    events = _provider.get_event(signature=signature)

    for event in events:

        if not event:
            return False, signature

        event_name = event.get("name")
        if event_name:
            log.info(
                "Event (signature: %s, name: %s) guessed from external source.",
                signature,
                event_name,
            )
            return True, event.get("name", signature)

    return False, signature


def _prepare_parameter_semantics(
    args, is_tuple: bool, unknown: bool
) -> List[ParameterSemantics]:
    if not args:
        return []

    elif not is_tuple:
        return [
            ParameterSemantics(
                parameter_name=arg.name if not unknown else f"arg_{i}",
                parameter_type=arg.type if not unknown else arg,
            )
            for i, arg in enumerate(args)
        ]

    return [
        ParameterSemantics(
            parameter_name="params",
            parameter_type="tuple",
            components=[
                ParameterSemantics(
                    parameter_name=arg.name if not unknown else f"arg_{i}",
                    parameter_type=arg.type if not unknown else arg,
                )
                for i, arg in enumerate(args)
            ],
        )
    ]



from ethtx.providers.node.connection_base import NodeConnection

CHAIN = "mainnet"
NODE = "test"
POA = False


class TestConnectionBase:
    @classmethod
    def setup_class(cls):
        cls.connection = NodeConnection(CHAIN, NODE, POA)

    def test_connection_parameters(self):
        assert self.connection.chain == CHAIN
        assert self.connection.url == NODE
        assert self.connection.poa == POA

    def test_connection_representation(self):
        representation = f"<Chain: {CHAIN}, Node: {NODE}, Poa: {POA}>"
        assert repr(self.connection) == representation
        assert str(self.connection) == representation

    def test_connection_dict(self):
        assert dict(self.connection) == {"chain": CHAIN, "url": NODE, "poa": POA}



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

import logging
from abc import ABC, abstractmethod
from typing import Any, TypeVar, Type

from ens import ENS
from web3 import Web3, exceptions

log = logging.getLogger(__name__)

T = TypeVar("T")


class ENSProviderBase(ABC):
    @abstractmethod
    def name(self, provider: Type[T], address: Any):
        ...

    @abstractmethod
    def address(self, provider: Type[T], name: Any):
        ...


class Web3ENSProvider(ENSProviderBase):
    ns: ENS

    def name(self, provider: Web3, address: str, *, strict: bool = True) -> str:
        ns = self._set_provider(provider)
        check_sum_address = Web3.toChecksumAddress(address)

        try:
            name = ns.name(address=check_sum_address, strict=strict)
        except exceptions.BadFunctionCallOutput:
            log.warning(
                "ENS name not found for address: %s. There is no code associated with this address.",
                address,
            )
            name = None

        if name:
            log.info("ENS resolved an address: %s to name: %s", address, name)

        return name if name else address

    def address(self, provider: Web3, name: str) -> str:
        ns = self._set_provider(provider)
        address = ns.address(name=name)
        if address:
            log.info("ENS resolved name: %s to address: %s", name, address)

        return address if address else name

    def _set_provider(self, provider: Web3) -> ENS:
        return ENS.fromWeb3(provider)


ENSProvider = Web3ENSProvider()



from decimal import Decimal

from ethtx.models.decoded_model import (
    AddressInfo,
    DecodedTransactionMetadata,
    Argument,
    DecodedEvent,
    DecodedCall,
    DecodedTransfer,
    DecodedBalance,
    Proxy,
    DecodedTransaction,
)
from tests.models.mock import DecodedModelMock, FAKE_TIME, ObjectModelMock


class TestDecodedModels:
    def test_address_info(self):
        ai = AddressInfo(address="address", name="name")

        assert ai.address == "address"
        assert ai.name == "name"
        assert ai.badge is None

    def test_decoded_transaction_metadata(self):
        dtm = DecodedTransactionMetadata(
            chain_id="mainnet",
            tx_hash="0x12345",
            block_number=1,
            block_hash="0x12345",
            timestamp=FAKE_TIME,
            gas_price=1,
            sender=DecodedModelMock.ADDRESS_INFO,
            receiver=DecodedModelMock.ADDRESS_INFO,
            tx_index=1,
            tx_value=2,
            gas_limit=3,
            gas_used=4,
            success=True,
        )

        assert dtm.chain_id == "mainnet"
        assert dtm.tx_hash == "0x12345"
        assert dtm.block_number == 1
        assert dtm.block_hash == "0x12345"
        assert dtm.timestamp == FAKE_TIME
        assert dtm.gas_price == 1
        assert dtm.sender == DecodedModelMock.ADDRESS_INFO
        assert dtm.receiver == DecodedModelMock.ADDRESS_INFO
        assert dtm.tx_index == 1
        assert dtm.tx_value == 2
        assert dtm.gas_limit == 3
        assert dtm.gas_used == 4
        assert dtm.success

    def test_argument(self):
        a = Argument(name="name", type="type", value=1)

        assert a.name == "name"
        assert a.type == "type"
        assert a.value == 1

    def test_decoded_event(self):
        de = DecodedEvent(
            chain_id="mainnet",
            tx_hash="0x12345",
            timestamp=FAKE_TIME,
            contract=DecodedModelMock.ADDRESS_INFO,
            index=1,
            event_signature="0x0bc2390103cdcea68787f9f22f8be92ccf20f5eae0bb850fbb70af78e366e4dd",
            event_name="WalletAddressesSet",
            parameters=[DecodedModelMock.ARGUMENT, DecodedModelMock.ARGUMENT],
        )

        assert de.chain_id == "mainnet"
        assert de.tx_hash == "0x12345"
        assert de.timestamp == FAKE_TIME
        assert de.contract == DecodedModelMock.ADDRESS_INFO
        assert de.index == 1
        assert de.call_id is None
        assert (
            de.event_signature
            == "0x0bc2390103cdcea68787f9f22f8be92ccf20f5eae0bb850fbb70af78e366e4dd"
        )
        assert de.event_name == "WalletAddressesSet"
        assert de.parameters == [DecodedModelMock.ARGUMENT, DecodedModelMock.ARGUMENT]
        assert not de.event_guessed

    def test_decoded_call(self):
        dc = DecodedCall(
            chain_id="mainnet",
            tx_hash="0x12345",
            timestamp=FAKE_TIME,
            call_type="call",
            from_address=DecodedModelMock.ADDRESS_INFO,
            to_address=DecodedModelMock.ADDRESS_INFO,
            value=15,
            function_signature="0x521f8bed",
            function_name="getAllOperator",
            arguments=[DecodedModelMock.ARGUMENT],
            outputs=[],
            gas_used=15,
            status=True,
            indent=1,
        )

        assert dc.chain_id == "mainnet"
        assert dc.tx_hash == "0x12345"
        assert dc.timestamp == FAKE_TIME
        assert dc.call_id is None
        assert dc.call_type == "call"
        assert dc.from_address == DecodedModelMock.ADDRESS_INFO
        assert dc.to_address == DecodedModelMock.ADDRESS_INFO
        assert dc.value == 15
        assert dc.function_signature == "0x521f8bed"
        assert dc.function_name == "getAllOperator"
        assert dc.arguments == [DecodedModelMock.ARGUMENT]
        assert dc.outputs == []
        assert dc.gas_used == 15
        assert dc.error is None
        assert dc.status
        assert dc.indent == 1
        assert dc.subcalls == []
        assert not dc.function_guessed

    def test_decoded_transfer(self):
        dt = DecodedTransfer(
            from_address=DecodedModelMock.ADDRESS_INFO,
            to_address=DecodedModelMock.ADDRESS_INFO,
            token_symbol="ts",
            value=Decimal(0.15),
        )

        assert dt.from_address == DecodedModelMock.ADDRESS_INFO
        assert dt.to_address == DecodedModelMock.ADDRESS_INFO
        assert dt.token_address is None
        assert dt.token_symbol == "ts"
        assert dt.token_standard is None
        assert dt.value == 0.15

    def test_decoded_balance(self):
        db = DecodedBalance(holder=DecodedModelMock.ADDRESS_INFO, tokens=[{}])

        assert db.holder == DecodedModelMock.ADDRESS_INFO
        assert db.tokens == [{}]

    def test_decoded_transaction(self):
        de = DecodedEvent(
            chain_id="mainnet",
            tx_hash="0x12345",
            timestamp=FAKE_TIME,
            contract=DecodedModelMock.ADDRESS_INFO,
            index=1,
            event_signature="0x0bc2390103cdcea68787f9f22f8be92ccf20f5eae0bb850fbb70af78e366e4dd",
            event_name="WalletAddressesSet",
            parameters=[DecodedModelMock.ARGUMENT, DecodedModelMock.ARGUMENT],
        )
        dc = DecodedCall(
            chain_id="mainnet",
            tx_hash="0x12345",
            timestamp=FAKE_TIME,
            call_type="call",
            from_address=DecodedModelMock.ADDRESS_INFO,
            to_address=DecodedModelMock.ADDRESS_INFO,
            value=15,
            function_signature="0x521f8bed",
            function_name="getAllOperator",
            arguments=[DecodedModelMock.ARGUMENT],
            outputs=[],
            gas_used=15,
            status=True,
            indent=1,
        )
        dt = DecodedTransfer(
            from_address=DecodedModelMock.ADDRESS_INFO,
            to_address=DecodedModelMock.ADDRESS_INFO,
            token_symbol="ts",
            value=0.15,
        )

        db = DecodedBalance(holder=DecodedModelMock.ADDRESS_INFO, tokens=[{}])

        t = DecodedTransaction(
            block_metadata=ObjectModelMock.BLOCK_METADATA,
            metadata=DecodedModelMock.DECODED_TRANSACTION_METADATA,
            events=[de],
            calls=dc,
            transfers=[dt],
            balances=[db],
        )

        assert t.block_metadata == ObjectModelMock.BLOCK_METADATA
        assert t.metadata == DecodedModelMock.DECODED_TRANSACTION_METADATA
        assert t.events == [de]
        assert t.calls == dc
        assert t.transfers == [dt]
        assert t.balances == [db]
        assert not t.status

    def test_proxy(self):
        p = Proxy(address="address", name="name", type="type")

        assert p.address == "address"
        assert p.name == "name"
        assert p.type == "type"
        assert p.semantics is None
        assert p.token is None



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from ethtx.models.decoded_model import DecodedTransactionMetadata, AddressInfo
from ethtx.models.objects_model import BlockMetadata
from .abc import SemanticSubmoduleAbc


class SemanticMetadataDecoder(SemanticSubmoduleAbc):
    """Semantic Metadata Decoder."""

    def decode(
        self,
        block_metadata: BlockMetadata,
        tx_metadata: DecodedTransactionMetadata,
        chain_id: str,
    ) -> DecodedTransactionMetadata:
        """Semantically decode metadata."""

        decoded_metadata = DecodedTransactionMetadata(
            chain_id=chain_id,
            tx_hash=tx_metadata.tx_hash,
            block_number=block_metadata.block_number,
            block_hash=block_metadata.block_hash,
            timestamp=block_metadata.timestamp,
            gas_price=tx_metadata.gas_price / 10**9,
            sender=AddressInfo(
                address=tx_metadata.from_address,
                name=self.repository.get_address_label(
                    chain_id, tx_metadata.from_address
                ),
                badge="sender",
            ),
            receiver=AddressInfo(
                address=tx_metadata.to_address,
                name=self.repository.get_address_label(
                    chain_id, tx_metadata.to_address
                ),
                badge="receiver",
            ),
            tx_index=tx_metadata.tx_index,
            tx_value=tx_metadata.tx_value,
            gas_limit=tx_metadata.gas_limit,
            gas_used=tx_metadata.gas_used,
            success=tx_metadata.success,
        )

        return decoded_metadata



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from typing import List

from ethtx.models.decoded_model import DecodedCall, DecodedTransfer, AddressInfo
from ethtx.utils.measurable import RecursionLimit
from .abc import ABISubmoduleAbc

RECURSION_LIMIT = 2000
ZERO_ADDRESS = "0x" + 40 * "0"


class ABITransfersDecoder(ABISubmoduleAbc):
    """Abi Transfers Decoder."""

    def decode(self, call: DecodedCall, events, proxies) -> List:
        """Decode transfers."""
        transfers = []

        def _transfers_calls(decoded_call):
            if decoded_call.status and decoded_call.value:
                transfers.append(
                    DecodedTransfer(
                        from_address=decoded_call.from_address,
                        to_address=decoded_call.to_address,
                        token_standard="ETH",
                        token_address=ZERO_ADDRESS,
                        token_symbol="ETH",
                        value=decoded_call.value,
                    )
                )
            if decoded_call.subcalls:
                for sub_call in decoded_call.subcalls:
                    _transfers_calls(sub_call)

        if call:
            with RecursionLimit(RECURSION_LIMIT):
                _transfers_calls(call)

        for event in events:

            # signatures of Transfer event valid for ERC20 and ERC721 and
            # TransferSingle for ERC1155
            if (
                    event.event_signature in
                    ("0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef",
                     "0xc3d58168c5ae7397731d063d5bbf3d657854427343f4c083240f7aacaa2d0f62")
            ):

                # Transfer event
                if event.event_signature == "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef":
                    from_address = '0x' + event.parameters[0].value[-40:]
                    to_address = '0x' + event.parameters[1].value[-40:]
                    token_id = event.parameters[2].value
                    value = event.parameters[2].value
                # TransferSingle event
                else:
                    from_address = '0x' + event.parameters[1].value[-40:]
                    to_address = '0x' + event.parameters[2].value[-40:]
                    token_id = event.parameters[3].value
                    value = event.parameters[4].value

                from_name = self._repository.get_address_label(
                    event.chain_id, from_address, proxies
                )
                to_name = self._repository.get_address_label(
                    event.chain_id, to_address, proxies
                )

                standard = self._repository.get_standard(
                    event.chain_id, event.contract.address
                )

                if event.event_signature == "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef" and \
                        (standard == "ERC20" or event.contract.address in proxies):

                    (
                        _,
                        token_symbol,
                        token_decimals,
                        _,
                    ) = self._repository.get_token_data(
                        event.chain_id, event.contract.address, proxies
                    )
                    try:
                        value = value / 10 ** token_decimals
                    except:
                        value = 0

                    transfers.append(
                        DecodedTransfer(
                            from_address=AddressInfo(
                                address=from_address, name=from_name
                            ),
                            to_address=AddressInfo(address=to_address, name=to_name),
                            token_standard=standard,
                            token_address=event.contract.address,
                            token_symbol=token_symbol,
                            value=value,
                        )
                    )
                else:

                    (
                        _,
                        token_symbol,
                        token_decimals,
                        _,
                    ) = self._repository.get_token_data(
                        event.chain_id, event.contract.address, proxies
                    )

                    if event.event_signature == "0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef":
                        value = 1
                    else:
                        value = int(value, 16) if type(value) == str else value

                    if token_symbol == 'Unknown':
                        token_symbol = 'NFT'

                    if len(str(token_id)) > 8:
                        token_symbol = (
                            f"{token_symbol} {str(token_id)[:6]}..."
                            f"{str(token_id)[-2:]}"
                        )
                    else:
                        token_symbol = f"{token_symbol} {token_id}"
                    token_address = f"{event.contract.address}?a={token_id}#inventory"
                    transfers.append(
                        DecodedTransfer(
                            from_address=AddressInfo(
                                address=from_address, name=from_name
                            ),
                            to_address=AddressInfo(address=to_address, name=to_name),
                            token_standard=standard,
                            token_address=token_address,
                            token_symbol=token_symbol,
                            value=value,
                        )
                    )

        return transfers



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from typing import Optional, List, Dict, Tuple

from ethtx.decoders.decoders.semantics import decode_events_and_functions
from ethtx.models.semantics_model import (
    AddressSemantics,
    ContractSemantics,
    ParameterSemantics,
    ERC20Semantics,
    TransformationSemantics,
    FunctionSemantics,
    EventSemantics,
    Signature,
    SignatureArg,
)
from ethtx.providers import EtherscanProvider, ENSProvider
from ethtx.providers.semantic_providers.database import ISemanticsDatabase
from ethtx.providers.web3_provider import NodeDataProvider
from ethtx.semantics.protocols_router import amend_contract_semantics
from ethtx.semantics.solidity.precompiles import precompiles
from ethtx.semantics.standards.erc20 import ERC20_FUNCTIONS, ERC20_EVENTS
from ethtx.semantics.standards.erc721 import ERC721_FUNCTIONS, ERC721_EVENTS


class SemanticsRepository:
    def __init__(
        self,
        database_connection: ISemanticsDatabase,
        etherscan_provider: EtherscanProvider,
        web3provider: NodeDataProvider,
        ens_provider: ENSProvider,
        refresh_ens: bool = True,
    ):
        self.database = database_connection
        self.etherscan = etherscan_provider
        self._web3provider = web3provider
        self._ens_provider = ens_provider
        self.refresh_ens = refresh_ens

        self._records: Optional[List] = None

    def record(self) -> None:
        """Records is an array used to hold semantics used in tx decing process.
        This recording is used just for logging"""
        self._records = []

    def end_record(self) -> List:
        tmp_records = self._records
        self._records = None
        return tmp_records

    def _read_stored_semantics(
        self, address: str, chain_id: str
    ) -> Optional[AddressSemantics]:

        if not address:
            return None

        raw_address_semantics = self.database.get_address_semantics(chain_id, address)

        if not raw_address_semantics:
            return None

        address_semantics = AddressSemantics.from_mongo_record(
            raw_address_semantics, self.database
        )

        if (
            self.refresh_ens
            and address_semantics.name == address_semantics.address
            and not raw_address_semantics["is_contract"]
        ):
            address_semantics.name = self._ens_provider.name(
                provider=self._web3provider._get_node_connection(chain_id),
                address=address,
            )

            if address_semantics.name != address_semantics.address:
                self.update_semantics(address_semantics)

        return address_semantics

    def _create_address_semantics(
        self, chain_id: str, address: str
    ) -> AddressSemantics:
        ZERO_HASH = "0xc5d2460186f7233c927e7db2dcc703c0e500b653ca82273b7bfad8045d85a470"

        # try to read the semantics form the Etherscan provider
        provider = self._web3provider
        code_hash = provider.get_code_hash(address, chain_id)

        if code_hash != ZERO_HASH:
            # smart contract
            raw_semantics, decoded = self.etherscan.contract.get_contract_abi(
                chain_id, address
            )
            if decoded and raw_semantics:
                # raw semantics received from Etherscan
                events, functions = decode_events_and_functions(raw_semantics["abi"])
                standard, standard_semantics = self._decode_standard_semantics(
                    address, raw_semantics["name"], events, functions
                )
                if standard == "ERC20":
                    erc20_semantics = standard_semantics
                else:
                    proxy_erc20 = provider.guess_erc20_proxy(address, chain_id)
                    if proxy_erc20:
                        erc20_semantics = ERC20Semantics(**proxy_erc20)
                    else:
                        erc20_semantics = None
                contract_semantics = ContractSemantics(
                    code_hash=code_hash,
                    name=raw_semantics["name"],
                    events=events,
                    functions=functions,
                    transformations={},
                )
                address_semantics = AddressSemantics(
                    chain_id=chain_id,
                    address=address,
                    name=raw_semantics["name"],
                    is_contract=True,
                    contract=contract_semantics,
                    standard=standard,
                    erc20=erc20_semantics,
                )

            else:
                # try to guess if the address is a toke
                potential_erc20_semantics = provider.guess_erc20_token(
                    address, chain_id
                )
                if potential_erc20_semantics:
                    standard = "ERC20"
                    erc20_semantics = ERC20Semantics(
                        name=potential_erc20_semantics["name"],
                        symbol=potential_erc20_semantics["symbol"],
                        decimals=potential_erc20_semantics["decimals"],
                    )
                else:
                    standard = None
                    erc20_semantics = None

                contract_semantics = ContractSemantics(
                    code_hash=code_hash, name=address
                )
                address_semantics = AddressSemantics(
                    chain_id=chain_id,
                    address=address,
                    name=address,
                    is_contract=True,
                    contract=contract_semantics,
                    standard=standard,
                    erc20=erc20_semantics,
                )

        else:
            # externally owned address
            contract_semantics = ContractSemantics(code_hash=ZERO_HASH, name="EOA")
            name = self._ens_provider.name(
                provider=self._web3provider._get_node_connection(chain_id),
                address=address,
            )
            address_semantics = AddressSemantics(
                chain_id=chain_id,
                address=address,
                name=name,
                is_contract=False,
                contract=contract_semantics,
            )

        return address_semantics

    def get_semantics(self, chain_id: str, address: str) -> Optional[AddressSemantics]:

        if not address:
            return None

        address_semantics = self._read_stored_semantics(address, chain_id)
        if not address_semantics:
            address_semantics = self._create_address_semantics(chain_id, address)
            self.update_semantics(address_semantics)

        # amend semantics with locally stored updates
        amend_contract_semantics(address_semantics.contract)

        if self._records is not None:
            self._records.append(address)

        return address_semantics

    def _decode_standard_semantics(
        self, address, name, events, functions: Dict[str, FunctionSemantics]
    ) -> Tuple[Optional[str], Optional[ERC20Semantics]]:
        standard = None
        standard_semantics = None

        if not address:
            return standard, standard_semantics

        if all(erc20_event in events for erc20_event in ERC20_EVENTS) and all(
            erc20_function in functions for erc20_function in ERC20_FUNCTIONS
        ):
            standard = "ERC20"
            try:
                provider = self._web3provider
                token_data = provider.get_erc20_token(address, name, functions)
                standard_semantics = ERC20Semantics(
                    name=token_data["name"],
                    symbol=token_data["symbol"],
                    decimals=token_data["decimals"],
                )
            except Exception:
                standard_semantics = ERC20Semantics(name=name, symbol=name, decimals=18)
        elif all(erc721_event in events for erc721_event in ERC721_EVENTS) and all(
            erc721_function in functions for erc721_function in ERC721_FUNCTIONS
        ):
            standard = "ERC721"
            standard_semantics = None

        return standard, standard_semantics

    def get_event_abi(self, chain_id, address, signature) -> Optional[EventSemantics]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        event_semantics = semantics.contract.events.get(signature)

        return event_semantics

    def get_transformations(
        self, chain_id, address, signature
    ) -> Optional[Dict[str, TransformationSemantics]]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        transformations = semantics.contract.transformations.get(signature)

        return transformations

    def get_anonymous_event_abi(self, chain_id, address) -> Optional[EventSemantics]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        event_semantics = None

        anonymous_events = {
            signature
            for signature, event in semantics.contract.events.items()
            if event.anonymous
        }
        if len(anonymous_events) == 1:
            event_signature = anonymous_events.pop()
            event_semantics = semantics.contract.events[event_signature]

        return event_semantics

    def get_function_abi(
        self, chain_id, address, signature
    ) -> Optional[FunctionSemantics]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        function_semantics = semantics.contract.functions.get(signature)

        return function_semantics

    def get_constructor_abi(self, chain_id, address) -> Optional[FunctionSemantics]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        constructor_semantics = semantics.contract.functions.get("constructor")

        if constructor_semantics:
            constructor_semantics.outputs.append(
                ParameterSemantics(
                    parameter_name="__create_output__",
                    parameter_type="ignore",
                    indexed=False,
                    dynamic=True,
                )
            )

        return constructor_semantics

    # do not cache with lru - uses unhashable arguments
    def get_address_label(self, chain_id, address, proxies=None) -> str:

        if not address:
            return ""

        if int(address, 16) in precompiles:
            contract_label = "Precompiled"
        else:
            semantics = self.get_semantics(chain_id, address)
            if semantics.erc20:
                contract_label = semantics.erc20.symbol
            elif proxies and address in proxies:
                contract_label = proxies[address].name
            else:
                contract_label = (
                    semantics.name if semantics and semantics.name else address
                )

        return contract_label

    def check_is_contract(self, chain_id, address) -> bool:

        if not address:
            return False

        semantics = self.get_semantics(chain_id, address)
        is_contract = semantics is not None and semantics.is_contract

        return is_contract

    def get_standard(self, chain_id, address) -> Optional[str]:

        if not address:
            return None

        semantics = self.get_semantics(chain_id, address)
        return semantics.standard

    def get_token_data(
        self, chain_id, address, proxies=None
    ) -> Tuple[Optional[str], Optional[str], Optional[int], Optional[str]]:

        if not address:
            return None, None, None, None

        semantics = self.get_semantics(chain_id, address)
        if semantics.erc20:
            token_name = semantics.erc20.name if semantics.erc20 else address
            token_symbol = semantics.erc20.symbol if semantics.erc20 else "Unknown"
            token_decimals = semantics.erc20.decimals if semantics.erc20 else 18
        elif proxies and address in proxies and proxies[address].token:
            token_name = proxies[address].token.name
            token_symbol = proxies[address].token.symbol
            token_decimals = proxies[address].token.decimals
        else:
            token_name = address
            token_symbol = "Unknown"
            token_decimals = 18

        return token_name, token_symbol, token_decimals, "ERC20"

    def update_address(self, chain_id, address, contract) -> Dict:

        updated_address = {"network": chain_id, "address": address, **contract}
        self.database.insert_address(address=updated_address, update_if_exist=True)

        return updated_address

    def update_semantics(self, semantics) -> None:

        if not semantics:
            return

        contract_id = self.database.insert_contract(
            contract=semantics.contract.dict(), update_if_exist=True
        )

        updated_address_semantics = semantics.copy()
        updated_address_semantics.contract = (
            updated_address_semantics.contract.code_hash
        )
        self.database.insert_address(
            address=updated_address_semantics.dict(), update_if_exist=True
        )

        if contract_id:
            self.insert_contract_signatures(semantics.contract)

    def insert_contract_signatures(self, contract_semantics: ContractSemantics) -> None:
        for _, v in contract_semantics.functions.items():

            if not v.signature.startswith("0x"):
                continue

            if v.inputs and v.inputs[0].parameter_type == "tuple":
                args = [
                    SignatureArg(name=param.parameter_name, type=param.parameter_type)
                    for param in v.inputs[0].components
                ]
            else:
                args = (
                    [
                        SignatureArg(
                            name=param.parameter_name, type=param.parameter_type
                        )
                        for param in v.inputs
                    ]
                    if v.inputs
                    else []
                )

            new_signature = Signature(
                signature_hash=v.signature, name=v.name, args=args
            )

            self.update_or_insert_signature(new_signature)

    def get_most_used_signature(self, signature_hash: str) -> Optional[Signature]:
        signatures = list(
            self.database.get_signature_semantics(signature_hash=signature_hash)
        )

        if signatures:
            most_common_signature = max(signatures, key=lambda x: x["count"])
            signature = Signature(
                signature_hash=most_common_signature["signature_hash"],
                name=most_common_signature["name"],
                args=most_common_signature["args"],
                count=most_common_signature["count"],
                tuple=most_common_signature["tuple"],
                guessed=most_common_signature["guessed"],
            )

            return signature

        return None

    def update_or_insert_signature(self, signature: Signature) -> None:
        signatures = self.database.get_signature_semantics(
            signature_hash=signature.signature_hash
        )
        for sig in signatures:
            if (
                signature.name == sig["name"]
                and signature.signature_hash == sig["signature_hash"]
                and len(signature.args) == len(sig["args"])
            ):
                if signature.args and any(
                    arg for arg in list(sig["args"][0].values()) if "arg" in arg
                ):
                    sig["guessed"] = signature.guessed
                    for index, argument in enumerate(sig["args"]):
                        argument["name"] = signature.args[index].name
                        argument["type"] = signature.args[index].type

                sig["count"] += 1
                self.database.insert_signature(signature=sig, update_if_exist=True)
                break

        else:
            self.database.insert_signature(signature=signature.dict())

    def delete_semantics(self, chain_id: str, addresses: List[str]):
        for address in addresses:
            self.database.delete_semantics_by_address(chain_id, address)



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from typing import Dict, List, Optional

from ethtx.decoders.semantic.abc import ISemanticDecoder
from ethtx.decoders.semantic.balances import SemanticBalancesDecoder
from ethtx.decoders.semantic.calls import SemanticCallsDecoder
from ethtx.decoders.semantic.events import SemanticEventsDecoder
from ethtx.decoders.semantic.metadata import SemanticMetadataDecoder
from ethtx.decoders.semantic.transfers import SemanticTransfersDecoder
from ethtx.models.decoded_model import (
    DecodedTransactionMetadata,
    DecodedTransaction,
    DecodedEvent,
    DecodedTransfer,
    DecodedBalance,
    DecodedCall,
    Proxy,
)
from ethtx.models.objects_model import BlockMetadata


class SemanticDecoder(ISemanticDecoder):
    def decode_transaction(
        self,
        block: BlockMetadata,
        transaction: DecodedTransaction,
        proxies: Dict[str, Proxy],
        chain_id: str,
    ) -> DecodedTransaction:
        transaction.metadata = self.decode_metadata(
            block, transaction.metadata, chain_id
        )
        transaction.events = self.decode_events(
            transaction.events, transaction.metadata, proxies
        )
        transaction.calls = self.decode_calls(
            transaction.calls, transaction.metadata, proxies
        )
        transaction.transfers = self.decode_transfers(
            transaction.transfers, transaction.metadata
        )
        transaction.balances = self.decode_balances(
            transaction.balances, transaction.metadata
        )

        return transaction

    def decode_metadata(
        self,
        block_metadata: BlockMetadata,
        tx_metadata: DecodedTransactionMetadata,
        chain_id: str,
    ) -> DecodedTransactionMetadata:
        return SemanticMetadataDecoder(repository=self.repository).decode(
            block_metadata=block_metadata, tx_metadata=tx_metadata, chain_id=chain_id
        )

    def decode_event(
        self,
        event: DecodedEvent,
        tx_metadata: DecodedTransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
    ) -> DecodedEvent:
        return SemanticEventsDecoder(repository=self.repository).decode(
            events=event, tx_metadata=tx_metadata, proxies=proxies or {}
        )

    def decode_events(
        self,
        events: List[DecodedEvent],
        tx_metadata: DecodedTransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
    ) -> List[DecodedEvent]:
        return SemanticEventsDecoder(repository=self.repository).decode(
            events=events, tx_metadata=tx_metadata, proxies=proxies or {}
        )

    def decode_calls(
        self,
        call: DecodedCall,
        tx_metadata: DecodedTransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
    ) -> DecodedCall:
        return SemanticCallsDecoder(repository=self.repository).decode(
            call=call, tx_metadata=tx_metadata, proxies=proxies or {}
        )

    def decode_call(
        self,
        call: DecodedCall,
        tx_metadata: DecodedTransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
    ) -> DecodedCall:
        return SemanticCallsDecoder(repository=self.repository).decode(
            call=call, tx_metadata=tx_metadata, proxies=proxies or {}
        )

    def decode_transfers(
        self, transfers: List[DecodedTransfer], tx_metadata: DecodedTransactionMetadata
    ) -> List[DecodedTransfer]:
        return SemanticTransfersDecoder(repository=self.repository).decode(
            transfers=transfers, tx_metadata=tx_metadata
        )

    def decode_balances(
        self, balances: List[DecodedBalance], tx_metadata: DecodedTransactionMetadata
    ) -> List[DecodedBalance]:
        return SemanticBalancesDecoder(repository=self.repository).decode(
            balances=balances, tx_metadata=tx_metadata
        )



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from web3 import Web3

from ethtx.utils.cache_tools import cache


@cache
def is_eip1969_proxy(chain, delegator, delegate):
    implementation_slot = hex(
        int(Web3.keccak(text="eip1967.proxy.implementation").hex(), 16) - 1
    )
    try:
        implementation = (
            "0x"
            + chain.eth.get_storage_at(
                Web3.toChecksumAddress(delegator), implementation_slot
            ).hex()[-40:]
        )
        return implementation == delegate
    except:
        return False


@cache
def is_eip1969_beacon_proxy(chain, delegator, delegate):
    ibeacon_abi = """[
                        {
                            "inputs": [],
                            "name": "implementation",
                            "outputs": [
                                {
                                    "internalType": "address",
                                    "name": "",
                                    "type": "address"
                                }
                            ],
                            "stateMutability": "view",
                            "type": "function"
                        }
                    ]"""

    beacon_slot = hex(int(Web3.keccak(text="eip1967.proxy.beacon").hex(), 16) - 1)
    try:
        beacon = (
            "0x"
            + chain.eth.get_storage_at(
                Web3.toChecksumAddress(delegator), beacon_slot
            ).hex()[-40:]
        )
        beacon = chain.eth.contract(
            address=Web3.toChecksumAddress(beacon), abi=ibeacon_abi
        )
        implementation = beacon.functions.implementation().call()
        return implementation == Web3.toChecksumAddress(delegate)
    except:
        return False



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from abc import ABC
from typing import Dict, Optional, Any, List


class ISemanticsDatabase(ABC):
    """Semantics Database. Represents raw interface required to be
    implemented by a database that provides persistent
    data about semantics"""

    def get_address_semantics(self, chain_id: str, address: str) -> Optional[Dict]:
        ...

    def get_contract_semantics(self, code_hash: str) -> Optional[Dict]:
        ...

    def get_signature_semantics(self, signature_hash: str) -> Optional[List[Dict]]:
        ...

    def insert_contract(self, contract: dict, update_if_exist: bool = False) -> Any:
        ...

    def insert_address(self, address: dict, update_if_exist: bool = False) -> Any:
        ...

    def insert_signature(self, signature, update_if_exist: bool = False) -> Any:
        ...

    def delete_semantics_by_address(self, chain_id: str, address: str) -> None:
        ...



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from __future__ import annotations

from datetime import datetime
from typing import Optional, List

from ethtx.models._types import THexBytes
from ethtx.models.base_model import BaseModel
from ethtx.models.objects_model import BlockMetadata, TransactionMetadata, Event, Call


class W3Block(BaseModel):
    chain_id: str
    difficulty: int
    extraData: THexBytes
    gasLimit: int
    gasUsed: int
    hash: THexBytes
    logsBloom: THexBytes
    miner: str
    nonce: THexBytes
    number: int
    parentHash: THexBytes
    receiptsRoot: THexBytes
    sha3Uncles: THexBytes
    size: int
    stateRoot: THexBytes
    timestamp: int
    totalDifficulty: int
    transactions: list
    transactionsRoot: THexBytes
    uncles: list

    def to_object(self) -> BlockMetadata:
        block_hash = self.hash.hex()
        timestamp = datetime.utcfromtimestamp(self.timestamp)
        parent_hash = self.parentHash.hex()
        miner = self.miner.lower()
        gas_limit = self.gasLimit
        gas_used = self.gasUsed
        tx_count = len(self.transactions)

        return BlockMetadata(
            block_number=self.number,
            block_hash=block_hash,
            timestamp=timestamp,
            parent_hash=parent_hash,
            miner=miner,
            gas_limit=gas_limit,
            gas_used=gas_used,
            tx_count=tx_count,
        )


class W3Transaction(BaseModel):
    chain_id: str
    blockHash: THexBytes
    blockNumber: int
    from_address: str
    gas: int
    gasPrice: int
    hash: THexBytes
    input: str
    nonce: int
    r: THexBytes
    s: THexBytes
    to: Optional[str]
    transactionIndex: int
    v: int
    value: int

    def to_object(self, w3receipt: W3Receipt) -> TransactionMetadata:
        tx_hash = self.hash.hex()
        block_number = self.blockNumber
        tx_index = self.transactionIndex
        from_address = self.from_address.lower()
        to_address = (
            self.to.lower()
            if self.to
            else w3receipt.contractAddress.lower()
            if w3receipt.contractAddress
            else None
        )
        tx_value = self.value
        gas_limit = self.gas
        gas_price = self.gasPrice
        gas_used = w3receipt.gasUsed
        success = w3receipt.status == 1

        return TransactionMetadata(
            tx_hash=tx_hash,
            block_number=block_number,
            tx_index=tx_index,
            from_address=from_address,
            to_address=to_address,
            tx_value=tx_value,
            gas_limit=gas_limit,
            gas_price=gas_price,
            gas_used=gas_used,
            success=success,
        )


class W3Receipt(BaseModel):
    tx_hash: str
    chain_id: str
    blockHash: THexBytes
    blockNumber: int
    contractAddress: Optional[str]
    cumulativeGasUsed: int
    from_address: str
    gasUsed: int
    logsBloom: THexBytes
    root: Optional[str]
    status: int
    to_address: Optional[str]
    transactionHash: THexBytes
    transactionIndex: int
    logs: list = []


class W3Log(BaseModel):
    tx_hash: str
    chain_id: str
    address: str
    blockHash: THexBytes
    blockNumber: int
    data: str
    logIndex: int
    removed: bool
    topics: List[THexBytes]
    transactionHash: THexBytes
    transactionIndex: int

    def to_object(self) -> Event:
        contract = self.address.lower()
        log_index = self.logIndex
        log_data = self.data
        topics = []

        for i, _ in enumerate(self.topics):
            topics.append(self.topics[i].hex())

        return Event(
            contract=contract, topics=topics, log_data=log_data, log_index=log_index
        )


class W3CallTree(BaseModel):
    tx_hash: str
    chain_id: str
    type: str
    from_address: str
    to_address: Optional[str]
    input: str
    output: str
    value: Optional[str]
    time: Optional[str]
    gas: Optional[str]
    gasUsed: Optional[str]
    error: Optional[str]
    calls: list = []

    def to_object(self) -> Call:
        from_address = self.from_address
        to_address = self.to_address
        call_value = int(self.value, 16) if self.value else 0
        call_type = self.type.lower()
        call_data = self.input
        return_value = self.output
        gas_used = int(self.gasUsed, 16) if self.gasUsed else None
        call_gas = int(self.gas, 16) if self.gas else None
        status = self.error is None
        error = self.error

        call = Call(
            call_type=call_type,
            from_address=from_address,
            to_address=to_address,
            call_value=call_value,
            call_data=call_data,
            return_value=return_value,
            gas_used=gas_used,
            call_gas=call_gas,
            status=status,
            error=error,
        )

        for child_call in self.calls:
            call.subcalls.append(child_call.to_object())

        return call



import datetime

import pytest

from ethtx.models.decoded_model import AddressInfo, Argument, DecodedTransactionMetadata
from ethtx.models.objects_model import BlockMetadata, TransactionMetadata, Call, Event
from ethtx.models.semantics_model import ParameterSemantics, ContractSemantics

FAKE_TIME = datetime.datetime(2020, 12, 25, 17, 5, 55)


@pytest.fixture
def patch_datetime_now(monkeypatch):
    class MyDatetime:
        @classmethod
        def now(cls):
            return FAKE_TIME

    monkeypatch.setattr(datetime, "datetime", MyDatetime)


class DecodedModelMock:
    ADDRESS_INFO: AddressInfo = AddressInfo(address="address", name="name")
    ARGUMENT: Argument = Argument(name="name", type="type", value=1)
    DECODED_TRANSACTION_METADATA: DecodedTransactionMetadata = (
        DecodedTransactionMetadata(
            tx_hash="0x",
            block_number=1,
            gas_price=2,
            from_address="0xa",
            to_address="0xb",
            tx_index=3,
            tx_value=4,
            gas_limit=5,
            gas_used=1,
            success=False,
        )
    )


class ObjectModelMock:
    BLOCK_METADATA: BlockMetadata = BlockMetadata(
        block_number=15,
        block_hash="0x1",
        timestamp=FAKE_TIME,
        parent_hash="0x",
        miner="miner",
        gas_limit=12,
        gas_used=1,
        tx_count=5,
    )

    TRANSACTION_METADATA: TransactionMetadata = TransactionMetadata(
        tx_hash="0x",
        block_number=1,
        gas_price=2,
        from_address="0xa",
        to_address="0xb",
        tx_index=3,
        tx_value=4,
        gas_limit=5,
        gas_used=1,
        success=False,
    )

    CALL: Call = Call(
        call_type="call",
        from_address="0xa",
        to_address="0xb",
        call_value=10,
        call_data="0x00000000000000000",
        return_value="0xeeee",
        status=True,
    )

    EVENT: Event = Event(contract="0x", topics=["", ""], log_index=1)


class SemanticModelMock:
    PARAMETER_SEMANTICS: ParameterSemantics = ParameterSemantics(
        parameter_name="name", parameter_type="type"
    )

    CONTRACT_SEMANTICS: ContractSemantics = ContractSemantics(
        code_hash="0x", name="name"
    )



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

import logging

from ..models.semantics_model import ContractSemantics
from ..semantics.router import Router

log = logging.getLogger(__name__)


def amend_contract_semantics(semantics: ContractSemantics, router_=Router()):
    if semantics.code_hash in router_:
        try:
            semantics_updates = router_[semantics.code_hash]
            if "name" in semantics_updates:
                semantics.name = semantics_updates["name"]
            if "events" in semantics_updates:
                for signature, event_semantics in semantics_updates["events"].items():
                    semantics.events[signature] = event_semantics
            if "functions" in semantics_updates:
                for signature, function_semantics in semantics_updates[
                    "functions"
                ].items():
                    semantics.functions[signature] = function_semantics
            if "transformations" in semantics_updates:
                for signature, transformation in semantics_updates[
                    "transformations"
                ].items():
                    semantics.transformations[signature] = transformation

        except Exception as exc:
            log.warning(
                "Semantics update load for %s failed.",
                semantics.code_hash,
                exc_info=exc,
            )



from ethtx.models.w3_model import W3Block, W3Transaction, W3Receipt, W3Log, W3CallTree


class TestW3Models:
    def test_w3_block(self, mocker):
        hex_bytes = mocker.patch(
            "hexbytes.main.HexBytes",
            return_value="0x",
            new_callable=mocker.PropertyMock,
        )
        wb = W3Block(
            chain_id="mainnet",
            difficulty=1,
            extraData=hex_bytes,
            gasLimit=100,
            gasUsed=20,
            hash=hex_bytes,
            logsBloom=hex_bytes,
            miner="miner",
            nonce=hex_bytes,
            number=10,
            parentHash=hex_bytes,
            receiptsRoot=hex_bytes,
            sha3Uncles=hex_bytes,
            size=10,
            stateRoot=hex_bytes,
            timestamp=123123123123,
            totalDifficulty=1,
            transactions=[],
            transactionsRoot=hex_bytes,
            uncles=[],
        )

        assert wb.chain_id == "mainnet"
        assert wb.difficulty == 1
        assert wb.extraData == hex_bytes
        assert wb.gasLimit == 100
        assert wb.gasUsed == 20
        assert wb.hash == hex_bytes
        assert wb.logsBloom == hex_bytes
        assert wb.miner == "miner"
        assert wb.nonce == hex_bytes
        assert wb.number == 10
        assert wb.parentHash == hex_bytes
        assert wb.receiptsRoot == hex_bytes
        assert wb.sha3Uncles == hex_bytes
        assert wb.size == 10
        assert wb.stateRoot == hex_bytes
        assert wb.timestamp == 123123123123
        assert wb.totalDifficulty == 1
        assert wb.transactions == []
        assert wb.transactionsRoot == hex_bytes
        assert wb.uncles == []

    def test_w3_transaction(self, mocker):
        hex_bytes = mocker.patch(
            "hexbytes.main.HexBytes",
            return_value="0x",
            new_callable=mocker.PropertyMock,
        )

        wt = W3Transaction(
            chain_id="mainnet",
            blockHash=hex_bytes,
            blockNumber=5,
            from_address="0x",
            gas=5,
            gasPrice=5,
            hash=hex_bytes,
            input="0x",
            nonce=5,
            r=hex_bytes,
            s=hex_bytes,
            to="0x",
            transactionIndex=5,
            v=5,
            value=5,
        )

        assert wt.chain_id == "mainnet"
        assert wt.blockHash == hex_bytes
        assert wt.blockNumber == 5
        assert wt.from_address == "0x"
        assert wt.gas == 5
        assert wt.gasPrice == 5
        assert wt.hash == hex_bytes
        assert wt.input == "0x"
        assert wt.nonce == 5
        assert wt.r == hex_bytes
        assert wt.s == hex_bytes
        assert wt.to == "0x"
        assert wt.transactionIndex == 5
        assert wt.v == 5
        assert wt.value == 5

    def test_w3_receipt(self, mocker):
        hex_bytes = mocker.patch(
            "hexbytes.main.HexBytes",
            return_value="0x",
            new_callable=mocker.PropertyMock,
        )

        wr = W3Receipt(
            tx_hash="0x",
            chain_id="0x",
            blockHash=hex_bytes,
            blockNumber=10,
            contractAddress="0x",
            cumulativeGasUsed=10,
            from_address="0x",
            gasUsed=10,
            logsBloom=hex_bytes,
            status=10,
            to_address="0x",
            transactionHash=hex_bytes,
            transactionIndex=10,
            logs=[],
        )

        assert wr.tx_hash == "0x"
        assert wr.chain_id == "0x"
        assert wr.blockHash == hex_bytes
        assert wr.blockNumber == 10
        assert wr.contractAddress == "0x"
        assert wr.cumulativeGasUsed == 10
        assert wr.from_address == "0x"
        assert wr.gasUsed == 10
        assert wr.logsBloom == hex_bytes
        assert wr.status == 10
        assert wr.to_address == "0x"
        assert wr.transactionHash == hex_bytes
        assert wr.transactionIndex == 10
        assert wr.logs == []
        assert wr.root is None

    def test_w3_log(self, mocker):
        hex_bytes = mocker.patch(
            "hexbytes.main.HexBytes",
            return_value="0x",
            new_callable=mocker.PropertyMock,
        )
        wl = W3Log(
            tx_hash="0x",
            chain_id="0x",
            address="0x",
            blockHash=hex_bytes,
            blockNumber=55,
            data="0x",
            logIndex=55,
            removed=False,
            topics=[hex_bytes],
            transactionHash=hex_bytes,
            transactionIndex=55,
        )

        assert wl.tx_hash == "0x"
        assert wl.chain_id == "0x"
        assert wl.address == "0x"
        assert wl.blockHash == hex_bytes
        assert wl.blockNumber == 55
        assert wl.data == "0x"
        assert wl.logIndex == 55
        assert not wl.removed
        assert wl.topics == [hex_bytes]
        assert wl.transactionHash == hex_bytes
        assert wl.transactionIndex == 55

    def test_w3_call_tree(self):
        wct = W3CallTree(
            tx_hash="0xdas",
            chain_id="0xdas",
            type="0xdas",
            from_address="0xdas",
            to_address="0xdas",
            input="0xdas",
            output="0xdas",
            calls=[],
        )

        assert wct.tx_hash == "0xdas"
        assert wct.chain_id == "0xdas"
        assert wct.type == "0xdas"
        assert wct.from_address == "0xdas"
        assert wct.to_address == "0xdas"
        assert wct.input == "0xdas"
        assert wct.output == "0xdas"
        assert wct.calls == []
        assert wct.value is None
        assert wct.time is None
        assert wct.gas is None
        assert wct.gasUsed is None
        assert wct.error is None



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from typing import Dict

from mongoengine import connect
from pymongo import MongoClient

from .decoders.abi.decoder import ABIDecoder
from .decoders.decoder_service import DecoderService
from .decoders.semantic.decoder import SemanticDecoder
from .models.decoded_model import Proxy, DecodedTransaction
from .models.objects_model import Call
from .providers import EtherscanProvider, Web3Provider, ENSProvider
from .providers.semantic_providers import (
    ISemanticsDatabase,
    SemanticsRepository,
    MongoSemanticsDatabase,
)
from .utils.validators import assert_tx_hash


class EthTxConfig:
    mongo_connection_string: str
    etherscan_api_key: str
    web3nodes: Dict[str, dict]
    etherscan_urls: Dict[str, str]
    default_chain: str

    def __init__(
        self,
        mongo_connection_string: str,
        web3nodes: Dict[str, dict],
        etherscan_api_key: str,
        etherscan_urls: Dict[str, str],
        default_chain: str = "mainnet",
    ):
        self.mongo_connection_string = mongo_connection_string
        self.etherscan_api_key = etherscan_api_key
        self.web3nodes = web3nodes
        self.default_chain = default_chain
        self.etherscan_urls = etherscan_urls


class EthTxDecoders:
    semantic_decoder: SemanticDecoder
    abi_decoder: ABIDecoder

    def __init__(self, decoder_service: DecoderService):
        self._decoder_service = decoder_service
        self.abi_decoder: ABIDecoder = decoder_service.abi_decoder
        self.semantic_decoder: SemanticDecoder = decoder_service.semantic_decoder

    def decode_transaction(
        self, tx_hash: str, chain_id: str = None, recreate_semantics: bool = False
    ) -> DecodedTransaction:
        assert_tx_hash(tx_hash)
        return self._decoder_service.decode_transaction(chain_id, tx_hash, recreate_semantics)

    def get_proxies(self, call_tree: Call, chain_id: str) -> Dict[str, Proxy]:
        delegations = self._decoder_service.get_delegations(call_tree)
        return self._decoder_service.get_proxies(delegations, chain_id)


class EthTxProviders:
    web3provider: Web3Provider
    etherscan_provider: EtherscanProvider
    ens_provider: ENSProvider

    def __init__(
        self,
        web3provider: Web3Provider,
        etherscan_provider: EtherscanProvider,
        ens_provider: ENSProvider,
    ):
        self.web3provider = web3provider
        self.etherscan_provider = etherscan_provider
        self.ens_provider = ens_provider


class EthTx:
    def __init__(
        self,
        default_chain: str,
        database: ISemanticsDatabase,
        web3provider: Web3Provider,
        etherscan_provider: EtherscanProvider,
        ens_provider: ENSProvider,
    ):
        self._default_chain = default_chain
        self._semantics_repository = SemanticsRepository(
            database_connection=database,
            etherscan_provider=etherscan_provider,
            web3provider=web3provider,
            ens_provider=ens_provider,
        )

        abi_decoder = ABIDecoder(self.semantics, self._default_chain)
        semantic_decoder = SemanticDecoder(self.semantics, self._default_chain)
        decoder_service = DecoderService(
            abi_decoder, semantic_decoder, web3provider, self._default_chain
        )
        self._decoders = EthTxDecoders(decoder_service=decoder_service)
        self._providers = EthTxProviders(
            web3provider=web3provider,
            etherscan_provider=etherscan_provider,
            ens_provider=ens_provider,
        )

    @staticmethod
    def initialize(config: EthTxConfig):
        mongo_client: MongoClient = connect(host=config.mongo_connection_string)
        repository = MongoSemanticsDatabase(db=mongo_client.get_database())

        web3provider = Web3Provider(
            nodes=config.web3nodes, default_chain=config.default_chain
        )
        etherscan_provider = EtherscanProvider(
            api_key=config.etherscan_api_key,
            nodes=config.etherscan_urls,
            default_chain_id=config.default_chain,
        )

        ens_provider = ENSProvider

        return EthTx(
            config.default_chain,
            repository,
            web3provider,
            etherscan_provider,
            ens_provider,
        )

    @property
    def decoders(self) -> EthTxDecoders:
        """EthTx Decoders."""
        return self._decoders

    @property
    def semantics(self) -> SemanticsRepository:
        """EthTx Semantics Repository."""
        return self._semantics_repository

    @property
    def providers(self) -> EthTxProviders:
        """EthTx Providers."""
        return self._providers

    @property
    def default_chain(self) -> str:
        """Default chain."""
        return self._default_chain

    @default_chain.setter
    def default_chain(self, chain: str) -> None:
        """Default chain setter."""
        self._default_chain = chain



import pytest

from ethtx.providers.node.connection_base import NodeConnection
from ethtx.providers.node.pool import NodeConnectionPool

MAINNET_CHAIN = {"mainnet": {"hook": "a", "poa": True}}
GOERLI_CHAIN = {"goerli": {"hook": "a, b, c", "poa": False}}

GOERLI_NODE = NodeConnection("goerli", "B", False)


class TestNodeConnectionPool:
    def test_number_of_connections(self):
        pool = NodeConnectionPool(nodes=MAINNET_CHAIN)
        assert len(pool) == 1, "Number of connections should equal 1."

    def test_add_connection(self):
        pool = NodeConnectionPool(nodes=MAINNET_CHAIN)
        pool.add_connection(connection=GOERLI_NODE)

        assert (
            GOERLI_NODE in pool.connections
        ), f"{GOERLI_NODE} should be in pool of connections."

    def test_get_connection(self):
        pool = NodeConnectionPool(nodes=MAINNET_CHAIN)
        pool.add_connection(connection=GOERLI_NODE)
        connection = pool.get_connection("goerli")

        assert (
            connection[0] == GOERLI_NODE
        ), f"{GOERLI_NODE} should be in pool of connections."

    def test_add_wrong_type_connection(self):
        pool = NodeConnectionPool(nodes=MAINNET_CHAIN)
        with pytest.raises(ValueError):
            pool.add_connection((1, 1, 1))



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

import io
import os
import subprocess
import sys
from shutil import rmtree

import toml
from setuptools import find_packages, setup, Command

# root dir
root = os.path.abspath(os.path.dirname(__file__))

# Package meta-data.
NAME = "EthTx"
DESCRIPTION = "EthTx transaction decoder."
URL = "https://github.com/EthTx/ethtx"
EMAIL = "karol@tfi.team, tomek@tfi.team, piotr.rudnik@tfi.team"
AUTHOR = "Karol Chojnowski, Tomasz Mierzwa, Piotr Rudnik"
REQUIRES_PYTHON = ">=3.7.0"

REQUIRED = []
REQUIRED_TEST = []

about = {
    "__version__": subprocess.check_output(
        ["git", "describe", "--tags"], universal_newlines=True
    ).strip()
}

try:
    with io.open(os.path.join(root, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


def load_requirements(fname):
    """Load requirements from file."""
    try:
        with open(fname, "r") as fh:
            pipfile = fh.read()
        pipfile_toml = toml.loads(pipfile)
    except FileNotFoundError:
        return []

    try:
        required_packages = pipfile_toml["packages"].items()
    except KeyError:
        return []

    return [f"{pkg}{ver}" if ver != "*" else pkg for pkg, ver in required_packages]


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(root, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag {0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# *************** INSTALL *****************
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    license="Apache-2.0 License",
    packages=find_packages(exclude=["tests"]),
    install_requires=load_requirements("Pipfile"),
    include_package_data=True,
    test_suite="tests",
    classifiers=[
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    # $ setup.py publish support.
    cmdclass={"upload": UploadCommand},
)



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.


__all__ = [
    "NodeConnectionException",
    "ProcessingException",
    "InvalidTransactionHash",
    "InvalidEtherscanReturnCodeException",
]

import json
from typing import Dict


class NodeConnectionException(Exception):
    """Node Connection Exception."""

    def __init__(self):
        super().__init__("Couldn't connect to node(s)")


class ProcessingException(Exception):
    """Processing Exception."""

    def __init__(self, msg):
        super().__init__("Exception processing: " + msg)


class InvalidTransactionHash(Exception):
    """Invalid Transaction Hash."""

    def __init__(self, tx_hash):
        super().__init__("Invalid transaction hash provided: " + tx_hash)


class InvalidEtherscanReturnCodeException(Exception):
    def __init__(self, returned_code: int, params: Dict = None):
        params_msg = " with params: " + json.dumps(params) if params else ""
        msg = f"Invalid status code for etherscan request: {returned_code} {params_msg}"
        super().__init__(msg)



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

import glob
import importlib
import os

from ..semantics.base import BaseType, Base


class Router:
    """
    Semantics router.
    Returns all objects withs semantics to include.
    """

    root_dir = os.path.dirname(__file__)
    root_module_name = ".".join(__name__.split(".")[:-1])

    def __new__(cls) -> BaseType:
        return cls._get_semantics()

    @classmethod
    def _get_semantics(cls) -> BaseType:
        """
        Get all available semantics.
        Match pattern:
            - .py file
            - object is a class type
            - object is a Base subclass
        """
        rv = {}
        files = (
            semantic
            for semantic in glob.iglob(cls.root_dir + "**/**", recursive=True)
            if os.path.isfile(semantic)
            and "__" not in semantic
            and semantic.endswith(".py")
        )

        for filename in files:
            filename = filename.replace("/", ".").replace(".py", "")
            imported_module = importlib.import_module(
                f"{cls.root_module_name}{filename.split(cls.root_module_name)[-1]}"
            )
            for item in dir(imported_module):
                obj = getattr(imported_module, item)
                if isinstance(obj, type) and issubclass(obj, Base) and obj != Base:
                    rv[obj.code_hash] = obj.contract_semantics

        return rv



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

from abc import ABC, abstractmethod
from typing import Optional, Any, List, Dict

from ethtx.models.decoded_model import DecodedCall, DecodedTransfer, Proxy
from ethtx.models.objects_model import (
    Block,
    Transaction,
    Call,
    Event,
    TransactionMetadata,
    BlockMetadata,
)
from ethtx.providers.semantic_providers import SemanticsRepository


class ABIBasic:
    def __init__(self, repository: SemanticsRepository, chain_id: str):
        self._default_chain = chain_id
        self._repository: SemanticsRepository = repository


class ABISubmoduleAbc(ABC, ABIBasic):
    @abstractmethod
    def decode(self, *args, **kwargs) -> Any:
        ...


class IABIDecoder(ABC, ABIBasic):
    def __init__(
        self,
        repository: SemanticsRepository,
        chain_id: str,
        strict: Optional[bool] = False,
    ):
        super().__init__(repository, chain_id)
        self.strict = strict

    @abstractmethod
    def decode_transaction(
        self, block: Block, transaction: Transaction, proxies: Dict[str, Proxy]
    ):
        ...

    @abstractmethod
    def decode_calls(
        self,
        call: Call,
        block: BlockMetadata,
        transaction: TransactionMetadata,
        proxies: Dict[str, Proxy],
    ) -> ABISubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_events(
        self,
        events: [Event],
        block: BlockMetadata,
        transaction: TransactionMetadata,
        proxies: Dict[str, Proxy],
    ) -> ABISubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_transfers(
        self, call: DecodedCall, events: [Event], proxies: Dict[str, Proxy]
    ) -> ABISubmoduleAbc.decode:
        ...

    @abstractmethod
    def decode_balances(
        self, transfers: List[DecodedTransfer]
    ) -> ABISubmoduleAbc.decode:
        ...



# Copyright 2021 DAI FOUNDATION (the original version https://github.com/daifoundation/ethtx_ce)
# Copyright 2021-2022 Token Flow Insights SA (modifications to the original software as recorded
# in the changelog https://github.com/EthTx/ethtx/blob/master/CHANGELOG.md)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at: http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
# The product contains trademarks and other branding elements of Token Flow Insights SA which are
# not licensed under the Apache 2.0 license. When using or reproducing the code, please remove
# the trademark and/or other branding elements.

import logging
from typing import Optional, Dict

from ethtx.models.decoded_model import DecodedCall, Proxy, AddressInfo
from ethtx.models.objects_model import Call, TransactionMetadata, BlockMetadata
from ethtx.semantics.solidity.precompiles import precompiles
from ethtx.semantics.standards.erc20 import ERC20_FUNCTIONS
from ethtx.semantics.standards.erc721 import ERC721_FUNCTIONS
from ethtx.utils.measurable import RecursionLimit
from .abc import ABISubmoduleAbc
from .helpers.utils import (
    decode_function_abi_with_external_source,
    decode_function_abi_with_repository,
    upsert_guessed_function_semantics,
)
from ..decoders.parameters import decode_function_parameters, decode_graffiti_parameters

log = logging.getLogger(__name__)

RECURSION_LIMIT = 2000


class ABICallsDecoder(ABISubmoduleAbc):
    """Abi Calls Decoder."""

    def decode(
        self,
        call: Call,
        block: BlockMetadata,
        transaction: TransactionMetadata,
        proxies: Optional[Dict[str, Proxy]] = None,
        chain_id: Optional[str] = None,
    ) -> Optional[DecodedCall]:
        """Decode call with sub_calls."""

        if not call:
            return None

        indent = 0
        status = True
        call_id = ""

        decoded_root_call = self.decode_call(
            call, block, transaction, call_id, indent, status, proxies or {}, chain_id
        )

        with RecursionLimit(RECURSION_LIMIT):
            calls_tree = self._decode_nested_calls(
                decoded_root_call,
                block,
                transaction,
                call.subcalls,
                indent,
                status,
                proxies,
                chain_id,
            )

        return calls_tree

    def decode_call(
        self,
        call: Call,
        block: BlockMetadata,
        transaction: TransactionMetadata,
        call_id: str = "",
        indent: int = 0,
        status: bool = True,
        proxies: Dict[str, Proxy] = None,
        chain_id: str = None,
    ) -> DecodedCall:
        """Decode single call."""

        guessed = False
        chain_id = chain_id or self._default_chain

        if call.call_data:
            function_signature = call.call_data[:10]
        else:
            function_signature = None

        from_name = self._repository.get_address_label(
            chain_id, call.from_address, proxies
        )

        to_name = self._repository.get_address_label(chain_id, call.to_address, proxies)

        if call.call_type == "selfdestruct":
            function_name = call.call_type
            function_input, function_output = [], []

        elif call.call_type == "create2":
            # ToDo: parse constructor
            # ToDo: force semantics reload

            # constructor_abi = self.repository.get_constructor_abi(call.chain_id, call.to_address)
            function_name = "new"
            function_input, function_output = [], []

        elif self._repository.check_is_contract(chain_id, call.to_address):
            standard = self._repository.get_standard(chain_id, call.to_address)

            function_abi = self._repository.get_function_abi(
                chain_id, call.to_address, function_signature
            )

            function_signature = call.call_data[:10] if call.call_data else ""

            if not function_abi and call.to_address in proxies:
                # try to find signature in delegate-called contracts
                for semantic in proxies[call.to_address].semantics:
                    function_abi = (
                        semantic.contract.functions[function_signature]
                        if function_signature in semantic.contract.functions
                        else None
                    )
                    if function_abi:
                        break

            if not function_abi:
                if standard == "ERC20":
                    # decode ERC20 calls if ABI for them is not defined
                    function_abi = ERC20_FUNCTIONS.get(function_signature)
                elif standard == "ERC721":
                    # decode ERC721 calls if ABI for them is not defined
                    function_abi = ERC721_FUNCTIONS.get(function_signature)

            function_name = function_abi.name if function_abi else function_signature

            function_input, function_output = decode_function_parameters(
                call.call_data, call.return_value, function_abi, call.status
            )

            if function_name.startswith("0x") and len(function_signature) > 2:
                repository_functions = [
                    decode_function_abi_with_repository(
                        function_signature, self._repository
                    )
                ]
                decoded_functions = (
                    decode_function_abi_with_external_source(
                        signature=function_signature
                    )
                    if repository_functions[0][1] is None
                    else repository_functions
                )
                for guessed, decoded_function in decoded_functions:
                    try:
                        function_abi = decoded_function
                        function_name = function_abi.name
                        function_input, function_output = decode_function_parameters(
                            call.call_data, call.return_value, function_abi, call.status
                        )
                    except Exception as e:
                        log.info(
                            "Skipping getting function from external source and trying to get next. Error: %s",
                            e,
                        )
                        continue
                    else:
                        break
                if repository_functions[0][1] is None and decoded_functions:
                    upsert_guessed_function_semantics(
                        function_signature, function_abi, self._repository
                    )

            if (
                not call.status
                and function_output
                and function_output[0].name == "Error"
            ):
                error_description = function_output.pop()
                call.error = f'Failed with "{error_description.value}"'

        elif call.to_address and int(call.to_address, 16) in precompiles:
            function_semantics = precompiles[int(call.to_address, 16)]
            function_name = function_semantics.name
            function_input, function_output = decode_function_parameters(
                call.call_data,
                call.return_value,
                function_semantics,
                call.status,
                strip_signature=False,
            )
        else:
            function_name = "fallback"
            function_input = decode_graffiti_parameters(call.call_data)
            function_output = []

        return DecodedCall(
            chain_id=chain_id,
            tx_hash=transaction.tx_hash,
            timestamp=block.timestamp,
            call_id=call_id,
            call_type=call.call_type,
            from_address=AddressInfo(address=call.from_address, name=from_name),
            to_address=AddressInfo(address=call.to_address, name=to_name),
            value=call.call_value / 10**18,
            function_signature=function_signature,
            function_name=function_name,
            arguments=function_input,
            outputs=function_output,
            gas_used=call.gas_used,
            error=call.error,
            status=status,
            indent=indent,
            function_guessed=guessed,
        )

    def _decode_nested_calls(
        self,
        call: DecodedCall,
        block: BlockMetadata,
        transaction: TransactionMetadata,
        sub_calls,
        indent,
        status,
        proxies,
        chain_id,
    ) -> DecodedCall:
        """Decode nested calls. Call may have sub_calls, if they exist, it will recursively process them."""
        for i, sub_call in enumerate(sub_calls):
            status = status and call.status

            sub_call_id = (
                "_".join([call.call_id, str(i).zfill(4)]) if call.call_id else str(i)
            )
            decoded = self.decode_call(
                sub_call,
                block,
                transaction,
                sub_call_id,
                indent + 1,
                status,
                proxies,
                chain_id,
            )
            call.subcalls.append(decoded)

            if sub_call.subcalls:
                self._decode_nested_calls(
                    decoded,
                    block,
                    transaction,
                    sub_call.subcalls,
                    indent + 1,
                    status,
                    proxies,
                    chain_id,
                )

        return call
